@@ -55,6 +55,10 @@ struct _MMSmsPropertiesClass {
 
 GType mm_sms_properties_get_type (void);
 
+#if GLIB_CHECK_VERSION(2, 44, 0)
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (MMSmsProperties, g_object_unref)
+#endif
+
 MMSmsProperties *mm_sms_properties_new (void);
 
 void mm_sms_properties_set_text                    (MMSmsProperties *self,
@@ -71,9 +75,13 @@ void mm_sms_properties_set_smsc                    (MMSmsProperties *self,
 void mm_sms_properties_set_validity_relative       (MMSmsProperties *self,
                                                     guint validity);
 void mm_sms_properties_set_class                   (MMSmsProperties *self,
-                                                    gint class);
+                                                    gint message_class);
 void mm_sms_properties_set_delivery_report_request (MMSmsProperties *self,
                                                     gboolean request);
+void mm_sms_properties_set_teleservice_id          (MMSmsProperties *self,
+                                                    MMSmsCdmaTeleserviceId teleservice_id);
+void mm_sms_properties_set_service_category        (MMSmsProperties *self,
+                                                    MMSmsCdmaServiceCategory service_category);
 
 const gchar  *mm_sms_properties_get_text                    (MMSmsProperties *self);
 const guint8 *mm_sms_properties_get_data                    (MMSmsProperties *self,
@@ -86,6 +94,8 @@ MMSmsValidityType mm_sms_properties_get_validity_type       (MMSmsProperties *se
 guint         mm_sms_properties_get_validity_relative       (MMSmsProperties *self);
 gint          mm_sms_properties_get_class                   (MMSmsProperties *self);
 gboolean      mm_sms_properties_get_delivery_report_request (MMSmsProperties *self);
+MMSmsCdmaTeleserviceId   mm_sms_properties_get_teleservice_id   (MMSmsProperties *self);
+MMSmsCdmaServiceCategory mm_sms_properties_get_service_category (MMSmsProperties *self);
 
 /*****************************************************************************/
 /* ModemManager/libmm-glib/mmcli specific methods */
@@ -99,8 +109,6 @@ MMSmsProperties *mm_sms_properties_new_from_string (const gchar *str,
 MMSmsProperties *mm_sms_properties_new_from_dictionary (GVariant *dictionary,
                                                         GError **error);
 
-MMSmsProperties *mm_sms_properties_dup (MMSmsProperties *orig);
-
 GVariant *mm_sms_properties_get_dictionary (MMSmsProperties *self);
 
 #endif
