@@ -23,6 +23,8 @@
 #include <string.h>
 
 #include <ModemManager.h>
+#include <ModemManager-tags.h>
+
 #include <mm-errors-types.h>
 #include <mm-gdbus-modem.h>
 
@@ -36,6 +38,10 @@
 
 G_DEFINE_ABSTRACT_TYPE (MMBaseModem, mm_base_modem, MM_GDBUS_TYPE_OBJECT_SKELETON);
 
+/* If we get 10 consecutive timeouts in a serial port, we consider the modem
+ * invalid and we request re-probing. */
+#define DEFAULT_MAX_TIMEOUTS 10
+
 enum {
     PROP_0,
     PROP_VALID,
@@ -89,6 +95,13 @@ struct _MMBaseModemPrivate {
     MMPortSerialAt *gps_control;
     MMPortSerialGps *gps;
 
+    /* Some audio-capable devices will have a port for audio specifically */
+    MMPortSerial *audio;
+
+    /* Support for parallel enable/disable operations */
+    GList *enable_tasks;
+    GList *disable_tasks;
+
 #if defined WITH_QMI
     /* QMI ports */
     GList *qmi;
@@ -191,6 +204,8 @@ mm_base_modem_grab_port (MMBaseModem         *self,
     }
     /* Serial ports... */
     else if (g_str_equal (subsys, "tty")) {
+        const gchar *flow_control_tag;
+
         if (ptype == MM_PORT_TYPE_QCDM)
             /* QCDM port */
             port = MM_PORT (mm_port_serial_qcdm_new (name));
@@ -203,11 +218,26 @@ mm_base_modem_grab_port (MMBaseModem         *self,
                                                    mm_serial_parser_v1_parse,
                                                    mm_serial_parser_v1_new (),
                                                    mm_serial_parser_v1_destroy);
-            /* Store flags already */
+            /* Prefer plugin-provided flags to the generic ones */
+            if (at_pflags == MM_PORT_SERIAL_AT_FLAG_NONE) {
+                if (mm_kernel_device_get_property_as_boolean (kernel_device, ID_MM_PORT_TYPE_AT_PRIMARY)) {
+                    mm_dbg ("AT port '%s/%s' flagged as primary", subsys, name);
+                    at_pflags = MM_PORT_SERIAL_AT_FLAG_PRIMARY;
+                } else if (mm_kernel_device_get_property_as_boolean (kernel_device, ID_MM_PORT_TYPE_AT_SECONDARY)) {
+                    mm_dbg ("AT port '%s/%s' flagged as secondary", subsys, name);
+                    at_pflags = MM_PORT_SERIAL_AT_FLAG_SECONDARY;
+                } else if (mm_kernel_device_get_property_as_boolean (kernel_device, ID_MM_PORT_TYPE_AT_PPP)) {
+                    mm_dbg ("AT port '%s/%s' flagged as PPP", subsys, name);
+                    at_pflags = MM_PORT_SERIAL_AT_FLAG_PPP;
+                }
+            }
             mm_port_serial_at_set_flags (MM_PORT_SERIAL_AT (port), at_pflags);
         } else if (ptype == MM_PORT_TYPE_GPS) {
             /* Raw GPS port */
             port = MM_PORT (mm_port_serial_gps_new (name));
+        } else if (ptype == MM_PORT_TYPE_AUDIO) {
+            /* Generic audio port */
+            port = MM_PORT (mm_port_serial_new (name, ptype));
         } else {
             g_set_error (error,
                          MM_CORE_ERROR,
@@ -226,11 +256,28 @@ mm_base_modem_grab_port (MMBaseModem         *self,
                               G_CALLBACK (serial_port_timed_out_cb),
                               self);
 
-        /* For serial ports, optionally use a specific baudrate */
-        if (mm_kernel_device_has_property (kernel_device, "ID_MM_TTY_BAUDRATE"))
+        /* For serial ports, optionally use a specific baudrate and flow control */
+        if (mm_kernel_device_has_property (kernel_device, ID_MM_TTY_BAUDRATE))
             g_object_set (port,
-                          MM_PORT_SERIAL_BAUD, mm_kernel_device_get_property_as_int (kernel_device, "ID_MM_TTY_BAUDRATE"),
+                          MM_PORT_SERIAL_BAUD, mm_kernel_device_get_property_as_int (kernel_device, ID_MM_TTY_BAUDRATE),
                           NULL);
+
+        flow_control_tag = mm_kernel_device_get_property (kernel_device, ID_MM_TTY_FLOW_CONTROL);
+        if (flow_control_tag) {
+            MMFlowControl flow_control;
+            GError *inner_error = NULL;
+
+            flow_control = mm_flow_control_from_string (flow_control_tag, &inner_error);
+            if (flow_control == MM_FLOW_CONTROL_UNKNOWN) {
+                mm_warn ("(%s/%s) unsupported flow control settings in port: %s",
+                         subsys, name, inner_error->message);
+                g_error_free (inner_error);
+            } else {
+                g_object_set (port,
+                              MM_PORT_SERIAL_FLOW_CONTROL, flow_control,
+                              NULL);
+            }
+        }
     }
     /* Net ports... */
     else if (g_str_equal (subsys, "net")) {
@@ -310,49 +357,123 @@ mm_base_modem_grab_port (MMBaseModem         *self,
 }
 
 gboolean
-mm_base_modem_disable_finish (MMBaseModem *self,
-                              GAsyncResult *res,
-                              GError **error)
+mm_base_modem_disable_finish (MMBaseModem   *self,
+                              GAsyncResult  *res,
+                              GError       **error)
 {
-    return MM_BASE_MODEM_GET_CLASS (self)->disable_finish (self, res, error);
+    return g_task_propagate_boolean (G_TASK (res), error);
+}
+
+static void
+disable_ready (MMBaseModem  *self,
+               GAsyncResult *res)
+{
+    GError *error = NULL;
+    GList  *l;
+    GList  *disable_tasks;
+
+    g_assert (self->priv->disable_tasks);
+    disable_tasks = self->priv->disable_tasks;
+    self->priv->disable_tasks = NULL;
+
+    MM_BASE_MODEM_GET_CLASS (self)->disable_finish (self, res, &error);
+    for (l = disable_tasks; l; l = g_list_next (l)) {
+        if (error)
+            g_task_return_error (G_TASK (l->data), g_error_copy (error));
+        else
+            g_task_return_boolean (G_TASK (l->data), TRUE);
+    }
+    g_clear_error (&error);
+
+    g_list_free_full (disable_tasks, g_object_unref);
 }
 
 void
-mm_base_modem_disable (MMBaseModem *self,
-                      GAsyncReadyCallback callback,
-                      gpointer user_data)
+mm_base_modem_disable (MMBaseModem         *self,
+                       GAsyncReadyCallback  callback,
+                       gpointer             user_data)
 {
+    GTask    *task;
+    gboolean  run_disable;
+
     g_assert (MM_BASE_MODEM_GET_CLASS (self)->disable != NULL);
     g_assert (MM_BASE_MODEM_GET_CLASS (self)->disable_finish != NULL);
 
+    /* If the list of disable tasks is empty, we need to run */
+    run_disable = !self->priv->disable_tasks;
+
+    /* Store task */
+    task = g_task_new (self, self->priv->cancellable, callback, user_data);
+    self->priv->disable_tasks = g_list_append (self->priv->disable_tasks, task);
+
+    if (!run_disable)
+        return;
+
     MM_BASE_MODEM_GET_CLASS (self)->disable (
         self,
         self->priv->cancellable,
-        callback,
-        user_data);
+        (GAsyncReadyCallback) disable_ready,
+        NULL);
 }
 
 gboolean
-mm_base_modem_enable_finish (MMBaseModem *self,
-                             GAsyncResult *res,
-                             GError **error)
+mm_base_modem_enable_finish (MMBaseModem   *self,
+                             GAsyncResult  *res,
+                             GError       **error)
 {
-    return MM_BASE_MODEM_GET_CLASS (self)->enable_finish (self, res, error);
+    return g_task_propagate_boolean (G_TASK (res), error);
+}
+
+static void
+enable_ready (MMBaseModem  *self,
+              GAsyncResult *res)
+{
+    GError *error = NULL;
+    GList  *l;
+    GList  *enable_tasks;
+
+    g_assert (self->priv->enable_tasks);
+    enable_tasks = self->priv->enable_tasks;
+    self->priv->enable_tasks = NULL;
+
+    MM_BASE_MODEM_GET_CLASS (self)->enable_finish (self, res, &error);
+    for (l = enable_tasks; l; l = g_list_next (l)) {
+        if (error)
+            g_task_return_error (G_TASK (l->data), g_error_copy (error));
+        else
+            g_task_return_boolean (G_TASK (l->data), TRUE);
+    }
+    g_clear_error (&error);
+
+    g_list_free_full (enable_tasks, g_object_unref);
 }
 
 void
-mm_base_modem_enable (MMBaseModem *self,
-                      GAsyncReadyCallback callback,
-                      gpointer user_data)
+mm_base_modem_enable (MMBaseModem         *self,
+                      GAsyncReadyCallback  callback,
+                      gpointer             user_data)
 {
+    GTask    *task;
+    gboolean  run_enable;
+
     g_assert (MM_BASE_MODEM_GET_CLASS (self)->enable != NULL);
     g_assert (MM_BASE_MODEM_GET_CLASS (self)->enable_finish != NULL);
 
+    /* If the list of enable tasks is empty, we need to run */
+    run_enable = !self->priv->enable_tasks;
+
+    /* Store task */
+    task = g_task_new (self, self->priv->cancellable, callback, user_data);
+    self->priv->enable_tasks = g_list_append (self->priv->enable_tasks, task);
+
+    if (!run_enable)
+        return;
+
     MM_BASE_MODEM_GET_CLASS (self)->enable (
         self,
         self->priv->cancellable,
-        callback,
-        user_data);
+        (GAsyncReadyCallback) enable_ready,
+        NULL);
 }
 
 gboolean
@@ -531,6 +652,22 @@ mm_base_modem_peek_port_gps (MMBaseModem *self)
     return self->priv->gps;
 }
 
+MMPortSerial *
+mm_base_modem_get_port_audio (MMBaseModem *self)
+{
+    g_return_val_if_fail (MM_IS_BASE_MODEM (self), NULL);
+
+    return (self->priv->audio ? g_object_ref (self->priv->audio) : NULL);
+}
+
+MMPortSerial *
+mm_base_modem_peek_port_audio (MMBaseModem *self)
+{
+    g_return_val_if_fail (MM_IS_BASE_MODEM (self), NULL);
+
+    return self->priv->audio;
+}
+
 #if defined WITH_QMI
 
 MMPortQmi *
@@ -814,6 +951,9 @@ mm_base_modem_get_port_infos (MMBaseModem *self,
         case MM_PORT_TYPE_GPS:
             port_infos[i].type = MM_MODEM_PORT_TYPE_GPS;
             break;
+        case MM_PORT_TYPE_AUDIO:
+            port_infos[i].type = MM_MODEM_PORT_TYPE_AUDIO;
+            break;
         case MM_PORT_TYPE_QMI:
             port_infos[i].type = MM_MODEM_PORT_TYPE_QMI;
             break;
@@ -844,6 +984,9 @@ mm_base_modem_find_ports (MMBaseModem *self,
     gpointer value;
     gpointer key;
 
+    if (!self->priv->ports)
+        return NULL;
+
     /* We'll iterate the ht of ports, looking for any port which is matches
      * the compare function */
     g_hash_table_iter_init (&iter, self->priv->ports);
@@ -921,6 +1064,7 @@ mm_base_modem_organize_ports (MMBaseModem *self,
     MMPortSerialQcdm *qcdm = NULL;
     MMPortSerialAt *gps_control = NULL;
     MMPortSerialGps *gps = NULL;
+    MMPortSerial *audio = NULL;
     MMPort *data_primary = NULL;
     GList *data = NULL;
 #if defined WITH_QMI
@@ -1011,6 +1155,12 @@ mm_base_modem_organize_ports (MMBaseModem *self,
                 gps = MM_PORT_SERIAL_GPS (candidate);
             break;
 
+        case MM_PORT_TYPE_AUDIO:
+            g_assert (MM_IS_PORT_SERIAL (candidate));
+            if (!audio)
+                audio = MM_PORT_SERIAL (candidate);
+            break;
+
 #if defined WITH_QMI
         case MM_PORT_TYPE_QMI:
             if (!qmi_primary)
@@ -1128,6 +1278,7 @@ mm_base_modem_organize_ports (MMBaseModem *self,
     log_port (self, MM_PORT (qcdm),         "qcdm");
     log_port (self, MM_PORT (gps_control),  "gps (control)");
     log_port (self, MM_PORT (gps),          "gps (nmea)");
+    log_port (self, MM_PORT (audio),        "audio");
 #if defined WITH_QMI
     log_port (self, MM_PORT (qmi_primary),  "qmi (primary)");
     for (l = qmi; l; l = g_list_next (l))
@@ -1318,6 +1469,8 @@ mm_base_modem_init (MMBaseModem *self)
                                                g_str_equal,
                                                g_free,
                                                g_object_unref);
+
+    self->priv->max_timeouts = DEFAULT_MAX_TIMEOUTS;
 }
 
 static void
@@ -1417,6 +1570,9 @@ finalize (GObject *object)
      * mm_auth_provider_cancel_for_owner (self->priv->authp, object);
     */
 
+    g_assert (!self->priv->enable_tasks);
+    g_assert (!self->priv->disable_tasks);
+
     mm_dbg ("Modem (%s) '%s' completely disposed",
             self->priv->plugin,
             self->priv->device);
@@ -1453,6 +1609,7 @@ dispose (GObject *object)
     g_clear_object (&self->priv->qcdm);
     g_clear_object (&self->priv->gps_control);
     g_clear_object (&self->priv->gps);
+    g_clear_object (&self->priv->audio);
 #if defined WITH_QMI
     /* We need to close the QMI port cleanly when disposing the modem object,
      * otherwise the allocated CIDs will be kept allocated, and if we end up
@@ -1497,7 +1654,7 @@ mm_base_modem_class_init (MMBaseModemClass *klass)
                            "Max timeouts",
                            "Maximum number of consecutive timed out commands sent to "
                            "the modem before disabling it. If 0, this feature is disabled.",
-                           0, G_MAXUINT, 0,
+                           0, G_MAXUINT, DEFAULT_MAX_TIMEOUTS,
                            G_PARAM_READWRITE);
     g_object_class_install_property (object_class, PROP_MAX_TIMEOUTS, properties[PROP_MAX_TIMEOUTS]);
 
