@@ -11,7 +11,8 @@
  * GNU General Public License for more details:
  *
  * Copyright (C) 2012 Google, Inc.
- * Copyright (C) 2012 Lanedo GmbH <aleksander@lanedo.com>
+ * Copyright (C) 2012 Lanedo GmbH
+ * Copyright (C) 2012-2019 Aleksander Morgado <aleksander@aleksander.es>
  */
 
 #include <ModemManager.h>
@@ -21,6 +22,7 @@
 #include "mm-iface-modem.h"
 #include "mm-iface-modem-location.h"
 #include "mm-log.h"
+#include "mm-modem-helpers.h"
 
 #define MM_LOCATION_GPS_REFRESH_TIME_SECS 30
 
@@ -139,7 +141,9 @@ build_location_dictionary (GVariant *previous,
                 break;
             case MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED:
                 g_assert_not_reached ();
-            case MM_MODEM_LOCATION_SOURCE_AGPS:
+            case MM_MODEM_LOCATION_SOURCE_AGPS_MSA:
+                g_assert_not_reached ();
+            case MM_MODEM_LOCATION_SOURCE_AGPS_MSB:
                 g_assert_not_reached ();
             default:
                 g_warn_if_reached ();
@@ -289,11 +293,12 @@ notify_3gpp_location_update (MMIfaceModemLocation *self,
 
     dbus_path = g_dbus_object_get_object_path (G_DBUS_OBJECT (self));
     mm_dbg ("Modem %s: 3GPP location updated "
-            "(MCC: '%u', MNC: '%u', Location area code: '%lX', Cell ID: '%lX')",
+            "(MCC: '%u', MNC: '%u', Location area code: '%lX', Tracking area code: '%lX', Cell ID: '%lX')",
             dbus_path,
             mm_location_3gpp_get_mobile_country_code (location_3gpp),
             mm_location_3gpp_get_mobile_network_code (location_3gpp),
             mm_location_3gpp_get_location_area_code (location_3gpp),
+            mm_location_3gpp_get_tracking_area_code (location_3gpp),
             mm_location_3gpp_get_cell_id (location_3gpp));
 
     /* We only update the property if we are supposed to signal
@@ -338,9 +343,10 @@ mm_iface_modem_location_3gpp_update_mcc_mnc (MMIfaceModemLocation *self,
 }
 
 void
-mm_iface_modem_location_3gpp_update_lac_ci (MMIfaceModemLocation *self,
-                                            gulong location_area_code,
-                                            gulong cell_id)
+mm_iface_modem_location_3gpp_update_lac_tac_ci (MMIfaceModemLocation *self,
+                                                gulong location_area_code,
+                                                gulong tracking_area_code,
+                                                gulong cell_id)
 {
     MmGdbusModemLocation *skeleton;
     LocationContext *ctx;
@@ -356,10 +362,9 @@ mm_iface_modem_location_3gpp_update_lac_ci (MMIfaceModemLocation *self,
         guint changed = 0;
 
         g_assert (ctx->location_3gpp != NULL);
-        changed += mm_location_3gpp_set_location_area_code (ctx->location_3gpp,
-                                                            location_area_code);
-        changed += mm_location_3gpp_set_cell_id (ctx->location_3gpp,
-                                                 cell_id);
+        changed += mm_location_3gpp_set_location_area_code (ctx->location_3gpp, location_area_code);
+        changed += mm_location_3gpp_set_tracking_area_code (ctx->location_3gpp, tracking_area_code);
+        changed += mm_location_3gpp_set_cell_id            (ctx->location_3gpp, cell_id);
         if (changed)
             notify_3gpp_location_update (self, skeleton, ctx->location_3gpp);
     }
@@ -381,14 +386,8 @@ mm_iface_modem_location_3gpp_clear (MMIfaceModemLocation *self)
         return;
 
     if (mm_gdbus_modem_location_get_enabled (skeleton) & MM_MODEM_LOCATION_SOURCE_3GPP_LAC_CI) {
-        guint changed = 0;
-
         g_assert (ctx->location_3gpp != NULL);
-        changed += mm_location_3gpp_set_location_area_code (ctx->location_3gpp, 0);
-        changed += mm_location_3gpp_set_cell_id (ctx->location_3gpp, 0);
-        changed += mm_location_3gpp_set_mobile_country_code (ctx->location_3gpp, 0);
-        changed += mm_location_3gpp_set_mobile_network_code (ctx->location_3gpp, 0);
-        if (changed)
+        if (mm_location_3gpp_reset (ctx->location_3gpp))
             notify_3gpp_location_update (self, skeleton, ctx->location_3gpp);
     }
 
@@ -510,7 +509,8 @@ update_location_source_status (MMIfaceModemLocation *self,
             g_clear_object (&ctx->location_cdma_bs);
         break;
     case MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED:
-    case MM_MODEM_LOCATION_SOURCE_AGPS:
+    case MM_MODEM_LOCATION_SOURCE_AGPS_MSA:
+    case MM_MODEM_LOCATION_SOURCE_AGPS_MSB:
         /* Nothing to setup in the context */
     default:
         break;
@@ -626,7 +626,7 @@ setup_gathering_step (GTask *task)
         return;
     }
 
-    while (ctx->current <= MM_MODEM_LOCATION_SOURCE_AGPS) {
+    while (ctx->current <= MM_MODEM_LOCATION_SOURCE_LAST) {
         gchar *source_str;
 
         if (ctx->to_enable & ctx->current) {
@@ -696,6 +696,7 @@ setup_gathering (MMIfaceModemLocation *self,
     MMModemLocationSource currently_enabled;
     MMModemLocationSource source;
     gchar *str;
+    gboolean allow_gps_unmanaged_always = FALSE;
 
     ctx = g_new (SetupGatheringContext, 1);
 
@@ -703,7 +704,8 @@ setup_gathering (MMIfaceModemLocation *self,
     g_task_set_task_data (task, ctx, (GDestroyNotify)setup_gathering_context_free);
 
     g_object_get (self,
-                  MM_IFACE_MODEM_LOCATION_DBUS_SKELETON, &ctx->skeleton,
+                  MM_IFACE_MODEM_LOCATION_DBUS_SKELETON,              &ctx->skeleton,
+                  MM_IFACE_MODEM_LOCATION_ALLOW_GPS_UNMANAGED_ALWAYS, &allow_gps_unmanaged_always,
                   NULL);
     if (!ctx->skeleton) {
         g_task_return_new_error (task,
@@ -722,8 +724,8 @@ setup_gathering (MMIfaceModemLocation *self,
     ctx->to_disable = MM_MODEM_LOCATION_SOURCE_NONE;
 
     /* Loop through all known bits in the bitmask to enable/disable specific location sources */
-    for (source = MM_MODEM_LOCATION_SOURCE_3GPP_LAC_CI;
-         source <= MM_MODEM_LOCATION_SOURCE_AGPS;
+    for (source = MM_MODEM_LOCATION_SOURCE_FIRST;
+         source <= MM_MODEM_LOCATION_SOURCE_LAST;
          source = source << 1) {
         /* skip unsupported sources */
         if (!(mm_gdbus_modem_location_get_capabilities (ctx->skeleton) & source))
@@ -749,13 +751,16 @@ setup_gathering (MMIfaceModemLocation *self,
     }
 
     /* When standard GPS retrieval (RAW/NMEA) is enabled, we cannot enable the
-     * UNMANAGED setup, and viceversa. */
-    if ((ctx->to_enable & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED &&
-         currently_enabled & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA)) ||
-        (ctx->to_enable & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA) &&
-         currently_enabled & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED) ||
-        (ctx->to_enable & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA) &&
-         ctx->to_enable & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED)) {
+     * UNMANAGED setup, and viceversa, unless explicitly allowed to do so by the
+     * plugin implementation (e.g. if the RAW/NMEA sources don't use the same TTY
+     * as the GPS UNMANAGED setup. */
+    if (!allow_gps_unmanaged_always &&
+        ((ctx->to_enable & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED &&
+          currently_enabled & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA)) ||
+         (ctx->to_enable & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA) &&
+          currently_enabled & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED) ||
+         (ctx->to_enable & (MM_MODEM_LOCATION_SOURCE_GPS_RAW | MM_MODEM_LOCATION_SOURCE_GPS_NMEA) &&
+          ctx->to_enable & MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED))) {
         g_task_return_new_error (task,
                                  MM_CORE_ERROR,
                                  MM_CORE_ERROR_FAILED,
@@ -764,6 +769,21 @@ setup_gathering (MMIfaceModemLocation *self,
         return;
     }
 
+    /* MSA A-GPS and MSB A-GPS cannot be set at the same time */
+    if ((ctx->to_enable & MM_MODEM_LOCATION_SOURCE_AGPS_MSA &&
+         currently_enabled & MM_MODEM_LOCATION_SOURCE_AGPS_MSB) ||
+        (ctx->to_enable & MM_MODEM_LOCATION_SOURCE_AGPS_MSB &&
+         currently_enabled & MM_MODEM_LOCATION_SOURCE_AGPS_MSA) ||
+        (ctx->to_enable & MM_MODEM_LOCATION_SOURCE_AGPS_MSA &&
+         ctx->to_enable & MM_MODEM_LOCATION_SOURCE_AGPS_MSB)) {
+        g_task_return_new_error (task,
+                                 MM_CORE_ERROR,
+                                 MM_CORE_ERROR_FAILED,
+                                 "Cannot have both MSA A-GPS and MSB A-GPS enabled at the same time");
+        g_object_unref (task);
+        return;
+    }
+
     if (ctx->to_enable != MM_MODEM_LOCATION_SOURCE_NONE) {
         str = mm_modem_location_source_build_string_from_mask (ctx->to_enable);
         mm_dbg ("Need to enable the following location sources: '%s'", str);
@@ -777,7 +797,7 @@ setup_gathering (MMIfaceModemLocation *self,
     }
 
     /* Start enabling/disabling location sources */
-    ctx->current = MM_MODEM_LOCATION_SOURCE_3GPP_LAC_CI;
+    ctx->current = MM_MODEM_LOCATION_SOURCE_FIRST;
     setup_gathering_step (task);
 }
 
@@ -981,7 +1001,7 @@ handle_set_supl_server_auth_ready (MMBaseModem *self,
     }
 
     /* If A-GPS is NOT supported, set error */
-    if (!(mm_gdbus_modem_location_get_capabilities (ctx->skeleton) & MM_MODEM_LOCATION_SOURCE_AGPS)) {
+    if (!(mm_gdbus_modem_location_get_capabilities (ctx->skeleton) & (MM_MODEM_LOCATION_SOURCE_AGPS_MSA | MM_MODEM_LOCATION_SOURCE_AGPS_MSB))) {
         g_dbus_method_invocation_return_error (ctx->invocation,
                                                MM_CORE_ERROR,
                                                MM_CORE_ERROR_UNSUPPORTED,
@@ -990,6 +1010,13 @@ handle_set_supl_server_auth_ready (MMBaseModem *self,
         return;
     }
 
+    /* Validate SUPL address string: either FQDN:PORT or IP:PORT */
+    if (!mm_parse_supl_address (ctx->supl, NULL, NULL, NULL, &error)) {
+        g_dbus_method_invocation_return_gerror (ctx->invocation, error);
+        handle_set_supl_server_context_free (ctx);
+        return;
+    }
+
     /* Check if plugin implements it */
     if (!MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->set_supl_server ||
         !MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->set_supl_server_finish) {
@@ -1032,6 +1059,108 @@ handle_set_supl_server (MmGdbusModemLocation *skeleton,
 
 /*****************************************************************************/
 
+typedef struct {
+    MmGdbusModemLocation  *skeleton;
+    GDBusMethodInvocation *invocation;
+    MMIfaceModemLocation  *self;
+    GVariant              *datav;
+} HandleInjectAssistanceDataContext;
+
+static void
+handle_inject_assistance_data_context_free (HandleInjectAssistanceDataContext *ctx)
+{
+    g_object_unref  (ctx->skeleton);
+    g_object_unref  (ctx->invocation);
+    g_object_unref  (ctx->self);
+    g_variant_unref (ctx->datav);
+    g_slice_free    (HandleInjectAssistanceDataContext, ctx);
+}
+
+static void
+inject_assistance_data_ready (MMIfaceModemLocation              *self,
+                              GAsyncResult                      *res,
+                              HandleInjectAssistanceDataContext *ctx)
+{
+    GError *error = NULL;
+
+    if (!MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->inject_assistance_data_finish (self, res, &error))
+        g_dbus_method_invocation_take_error (ctx->invocation, error);
+    else
+        mm_gdbus_modem_location_complete_inject_assistance_data (ctx->skeleton, ctx->invocation);
+
+    handle_inject_assistance_data_context_free (ctx);
+}
+
+static void
+handle_inject_assistance_data_auth_ready (MMBaseModem                       *self,
+                                          GAsyncResult                      *res,
+                                          HandleInjectAssistanceDataContext *ctx)
+{
+    GError       *error = NULL;
+    const guint8 *data;
+    gsize         data_size;
+
+    if (!mm_base_modem_authorize_finish (self, res, &error)) {
+        g_dbus_method_invocation_take_error (ctx->invocation, error);
+        handle_inject_assistance_data_context_free (ctx);
+        return;
+    }
+
+    /* If the type is NOT supported, set error */
+    if (mm_gdbus_modem_location_get_supported_assistance_data (ctx->skeleton) == MM_MODEM_LOCATION_ASSISTANCE_DATA_TYPE_NONE) {
+        g_dbus_method_invocation_return_error (ctx->invocation,
+                                               MM_CORE_ERROR,
+                                               MM_CORE_ERROR_UNSUPPORTED,
+                                               "Cannot inject assistance data: ununsupported");
+        handle_inject_assistance_data_context_free (ctx);
+        return;
+    }
+
+    /* Check if plugin implements it */
+    if (!MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->inject_assistance_data ||
+        !MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->inject_assistance_data_finish) {
+        g_dbus_method_invocation_return_error (ctx->invocation,
+                                               MM_CORE_ERROR,
+                                               MM_CORE_ERROR_UNSUPPORTED,
+                                               "Cannot inject assistance data: not implemented");
+        handle_inject_assistance_data_context_free (ctx);
+        return;
+    }
+
+    data = (const guint8 *) g_variant_get_fixed_array (ctx->datav, &data_size, sizeof (guint8));
+
+    /* Request to inject assistance data */
+    MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->inject_assistance_data (ctx->self,
+                                                                          data,
+                                                                          data_size,
+                                                                          (GAsyncReadyCallback)inject_assistance_data_ready,
+                                                                          ctx);
+}
+
+static gboolean
+handle_inject_assistance_data (MmGdbusModemLocation  *skeleton,
+                               GDBusMethodInvocation *invocation,
+                               GVariant              *datav,
+                               MMIfaceModemLocation  *self)
+{
+    HandleInjectAssistanceDataContext *ctx;
+
+    ctx = g_slice_new (HandleInjectAssistanceDataContext);
+    ctx->skeleton   = g_object_ref (skeleton);
+    ctx->invocation = g_object_ref (invocation);
+    ctx->self       = g_object_ref (self);
+    ctx->datav      = g_variant_ref (datav);
+
+    mm_base_modem_authorize (MM_BASE_MODEM (self),
+                             invocation,
+                             MM_AUTHORIZATION_DEVICE_CONTROL,
+                             (GAsyncReadyCallback)handle_inject_assistance_data_auth_ready,
+                             ctx);
+    return TRUE;
+}
+
+/*****************************************************************************/
+
 typedef struct {
     MmGdbusModemLocation *skeleton;
     GDBusMethodInvocation *invocation;
@@ -1386,7 +1515,8 @@ interface_enabling_step (GTask *task)
         default_sources &= ~(MM_MODEM_LOCATION_SOURCE_GPS_RAW |
                              MM_MODEM_LOCATION_SOURCE_GPS_NMEA |
                              MM_MODEM_LOCATION_SOURCE_GPS_UNMANAGED |
-                             MM_MODEM_LOCATION_SOURCE_AGPS);
+                             MM_MODEM_LOCATION_SOURCE_AGPS_MSA |
+                             MM_MODEM_LOCATION_SOURCE_AGPS_MSB);
 
         setup_gathering (self,
                          default_sources,
@@ -1445,6 +1575,8 @@ typedef enum {
     INITIALIZATION_STEP_CAPABILITIES,
     INITIALIZATION_STEP_VALIDATE_CAPABILITIES,
     INITIALIZATION_STEP_SUPL_SERVER,
+    INITIALIZATION_STEP_SUPPORTED_ASSISTANCE_DATA,
+    INITIALIZATION_STEP_ASSISTANCE_DATA_SERVERS,
     INITIALIZATION_STEP_GPS_REFRESH_RATE,
     INITIALIZATION_STEP_LAST
 } InitializationStep;
@@ -1462,6 +1594,55 @@ initialization_context_free (InitializationContext *ctx)
     g_free (ctx);
 }
 
+static void
+load_assistance_data_servers_ready (MMIfaceModemLocation *self,
+                                    GAsyncResult         *res,
+                                    GTask                *task)
+{
+    GError                 *error = NULL;
+    InitializationContext  *ctx;
+    gchar                 **servers;
+
+    ctx = g_task_get_task_data (task);
+
+    servers = MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_assistance_data_servers_finish (self, res, &error);
+    if (error) {
+        mm_warn ("couldn't load assistance data servers: '%s'", error->message);
+        g_error_free (error);
+    }
+
+    mm_gdbus_modem_location_set_assistance_data_servers (ctx->skeleton, (const gchar *const *)servers);
+    g_strfreev (servers);
+
+    /* Go on to next step */
+    ctx->step++;
+    interface_initialization_step (task);
+}
+
+static void
+load_supported_assistance_data_ready (MMIfaceModemLocation *self,
+                                      GAsyncResult         *res,
+                                      GTask                *task)
+{
+    GError                            *error = NULL;
+    MMModemLocationAssistanceDataType  mask;
+    InitializationContext             *ctx;
+
+    ctx = g_task_get_task_data (task);
+
+    mask = MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supported_assistance_data_finish (self, res, &error);
+    if (error) {
+        mm_warn ("couldn't load supported assistance data types: '%s'", error->message);
+        g_error_free (error);
+    }
+
+    mm_gdbus_modem_location_set_supported_assistance_data (ctx->skeleton, (guint32) mask);
+
+    /* Go on to next step */
+    ctx->step++;
+    interface_initialization_step (task);
+}
+
 static void
 load_supl_server_ready (MMIfaceModemLocation *self,
                         GAsyncResult *res,
@@ -1562,7 +1743,8 @@ interface_initialization_step (GTask *task)
 
     case INITIALIZATION_STEP_SUPL_SERVER:
         /* If the modem supports A-GPS, load SUPL server */
-        if (ctx->capabilities & MM_MODEM_LOCATION_SOURCE_AGPS &&
+        if ((ctx->capabilities & (MM_MODEM_LOCATION_SOURCE_AGPS_MSA |
+                                  MM_MODEM_LOCATION_SOURCE_AGPS_MSB)) &&
             MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supl_server &&
             MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supl_server_finish) {
             MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supl_server (
@@ -1574,6 +1756,37 @@ interface_initialization_step (GTask *task)
         /* Fall down to next step */
         ctx->step++;
 
+    case INITIALIZATION_STEP_SUPPORTED_ASSISTANCE_DATA:
+        /* If the modem supports any GPS-related technology, check assistance data types supported */
+        if ((ctx->capabilities & (MM_MODEM_LOCATION_SOURCE_AGPS_MSA |
+                                  MM_MODEM_LOCATION_SOURCE_AGPS_MSB |
+                                  MM_MODEM_LOCATION_SOURCE_GPS_RAW |
+                                  MM_MODEM_LOCATION_SOURCE_GPS_NMEA)) &&
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supported_assistance_data &&
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supported_assistance_data_finish) {
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_supported_assistance_data (
+                self,
+                (GAsyncReadyCallback)load_supported_assistance_data_ready,
+                task);
+            return;
+        }
+        /* Fall down to next step */
+        ctx->step++;
+
+    case INITIALIZATION_STEP_ASSISTANCE_DATA_SERVERS:
+        /* If any assistance data supported, load servers */
+        if ((mm_gdbus_modem_location_get_supported_assistance_data (ctx->skeleton) != MM_MODEM_LOCATION_ASSISTANCE_DATA_TYPE_NONE) &&
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_assistance_data_servers &&
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_assistance_data_servers_finish) {
+            MM_IFACE_MODEM_LOCATION_GET_INTERFACE (self)->load_assistance_data_servers (
+                self,
+                (GAsyncReadyCallback)load_assistance_data_servers_ready,
+                task);
+            return;
+        }
+        /* Fall down to next step */
+        ctx->step++;
+
     case INITIALIZATION_STEP_GPS_REFRESH_RATE:
         /* If we have GPS capabilities, expose the GPS refresh rate */
         if (ctx->capabilities & ((MM_MODEM_LOCATION_SOURCE_GPS_RAW |
@@ -1596,6 +1809,10 @@ interface_initialization_step (GTask *task)
                           "handle-set-supl-server",
                           G_CALLBACK (handle_set_supl_server),
                           self);
+        g_signal_connect (ctx->skeleton,
+                          "handle-inject-assistance-data",
+                          G_CALLBACK (handle_inject_assistance_data),
+                          self);
         g_signal_connect (ctx->skeleton,
                           "handle-set-gps-refresh-rate",
                           G_CALLBACK (handle_set_gps_refresh_rate),
@@ -1644,6 +1861,7 @@ mm_iface_modem_location_initialize (MMIfaceModemLocation *self,
 
         /* Set all initial property defaults */
         mm_gdbus_modem_location_set_capabilities (skeleton, MM_MODEM_LOCATION_SOURCE_NONE);
+        mm_gdbus_modem_location_set_supported_assistance_data (skeleton, MM_MODEM_LOCATION_ASSISTANCE_DATA_TYPE_NONE);
         mm_gdbus_modem_location_set_enabled (skeleton, MM_MODEM_LOCATION_SOURCE_NONE);
         mm_gdbus_modem_location_set_signals_location (skeleton, FALSE);
         mm_gdbus_modem_location_set_location (skeleton,
@@ -1696,6 +1914,14 @@ iface_modem_location_init (gpointer g_iface)
                               MM_GDBUS_TYPE_MODEM_LOCATION_SKELETON,
                               G_PARAM_READWRITE));
 
+    g_object_interface_install_property
+        (g_iface,
+         g_param_spec_boolean (MM_IFACE_MODEM_LOCATION_ALLOW_GPS_UNMANAGED_ALWAYS,
+                               "Allow unmanaged GPS always",
+                               "Whether to always allow GPS unmanaged, even when raw/nmea GPS sources are enabled",
+                               FALSE,
+                               G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
     initialized = TRUE;
 }
 
